/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2016-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"reflect"
	"strconv"
	"strings"
)

type Features struct {
	// 'BlueprintEmbed' is a special case in Blueprint which makes it interpret
	// a runtime-generated type as being embedded in its parent struct.
	BlueprintEmbed interface{}
}

type ConfigPropertiesJson struct {
	// Map of all available features (e.g. noasserts: { cflags: ["-DNDEBUG]" }),
	// and whether they are enabled or not.
	Features map[string]bool
	// Map of all available properties which can be used in templates. Features are
	// not automatically included in this by Bob, so should be added explicitly by the
	// config system if required. These are converted to strings, then made available
	// for use in templates.
	Properties map[string]interface{}
}

type ConfigProperties struct {
	ConfigPropertiesJson
	stringMap map[string]string
}

func (properties ConfigProperties) getProp(name string) interface{} {
	if elem, ok := properties.Properties[name]; ok {
		return elem
	}
	panic(fmt.Sprintf("No property found: %s", name))
}

func (properties ConfigProperties) GetBool(name string) bool {
	if ret, ok := properties.getProp(name).(bool); ok {
		return ret
	}
	panic(fmt.Sprintf("Property %s is not a bool", name))
}

func (properties ConfigProperties) GetString(name string) string {
	if ret, ok := properties.getProp(name).(string); ok {
		return ret
	}
	panic(fmt.Sprintf("Property %s is not a string", name))
}

func (properties ConfigProperties) StringMap() map[string]string {
	return properties.stringMap
}

func LoadConfig(filename string) *ConfigProperties {
	properties := &ConfigProperties{}
	content, err := ioutil.ReadFile(filename)
	if err != nil {
		panic(err)
	}
	d := json.NewDecoder(bytes.NewReader(content))

	// Decode numbers in JSON as json.Numbers instead of float64.
	// This is actually a string, which is what we want.
	d.UseNumber()
	err = d.Decode(&properties.ConfigPropertiesJson)
	if err != nil {
		panic(err)
	}

	properties.stringMap = make(map[string]string)
	for key, val := range properties.Properties {
		properties.stringMap[key] = convertToString(val)
	}

	return properties
}

// This function converts a config value into a string, using the following rules:
//  - booleans are converted into "0" or "1"
//  - Strings are used as-is
//  - Ints are converted into 10-base form
//  - Slices of booleans,strings and ints are converted into a space-separated string
//  - Pointers to booleans,strings and ints are converted into the referenced value
//
// Any other type might panic.
func convertToString(thing interface{}) string {
	field := reflect.ValueOf(thing)
	var value string
	switch field.Kind() {
	case reflect.String:
		value = field.String()

	case reflect.Bool:
		if field.Bool() {
			value = "1"
		} else {
			value = "0"
		}

	case reflect.Int:
		value = strconv.FormatInt(field.Int(), 10)

	case reflect.Ptr:
		if !reflect.Indirect(field).IsValid() {
			// This happens if we have nil pointer. The only time this happens
			// is if we have a "special" boolean.  Ignore these for now.
		} else {
			value = convertToString(reflect.Indirect(field))
		}

	case reflect.Slice:
		values := []string{}
		for j := 0; j < field.Len(); j++ {
			elem := field.Index(j)
			values = append(values, convertToString(elem))
		}
		value = strings.Join(values, " ")

	default:
		panic(fmt.Sprintf("Can't convert type %s to string!", field.Type().String()))
	}
	return value
}
