/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2017-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"fmt"
	"path/filepath"
	"reflect"
	"strings"

	"github.com/google/blueprint"
)

type KernelModule struct {
	blueprint.SimpleName
	Properties struct {
		Features
		Build
		Defaults []string
	}
}

func (m *KernelModule) Defaults() []string {
	return m.Properties.Defaults
}

func (m *KernelModule) Build() *Build {
	return &m.Properties.Build
}

func (m *KernelModule) topLevelProperties() []interface{} {
	return []interface{}{&m.Properties.Build.BuildProps}
}

func (m *KernelModule) Features() *Features {
	return &m.Properties.Features
}

func (m *KernelModule) OutputName() string {
	if len(m.Properties.Out) > 0 {
		return m.Properties.Out
	}
	return m.SimpleName.Name()
}

func (m *KernelModule) AltName() string {
	return m.OutputName()
}

func (m *KernelModule) AltShortName() string {
	return m.AltName()
}

func (m *KernelModule) ShortName() string {
	return m.Name()
}

func (m *KernelModule) GetEnableableProps() *EnableableProps {
	return &m.Properties.EnableableProps
}

func (m *KernelModule) OutputDir(g GeneratorBackend) string {
	return g.kernelModOutputDir(m)
}

func (m *KernelModule) Outputs(g GeneratorBackend) []string {
	return []string{filepath.Join(m.OutputDir(g), m.OutputName()+".ko")}
}

func (m *KernelModule) FilesToInstall(ctx blueprint.ModuleContext) []string {
	return m.Outputs(getBackend(ctx))
}

func (m *KernelModule) GetInstallableProps() *InstallableProps {
	return &m.Properties.InstallableProps
}

func (m *KernelModule) GetInstallDepPhonyNames(ctx blueprint.ModuleContext) []string {
	return GetShortNamesForDirectDepsWithTags(ctx, InstallDepTag, KernelModuleDepTag)
}

func (m *KernelModule) processPaths(ctx blueprint.BaseModuleContext) {
	m.Properties.Build.processPaths(ctx)
}

func (m *KernelModule) extraSymbolsFiles(ctx blueprint.ModuleContext) (files []string) {
	g := getBackend(ctx)

	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool { return ctx.OtherModuleDependencyTag(m) == KernelModuleDepTag },
		func(m blueprint.Module) {
			if km, ok := m.(*KernelModule); ok {
				file := filepath.Join(km.OutputDir(g), "Module.symvers")
				files = append(files, file)
			} else {
				panic(fmt.Errorf("invalid extra_symbols, %s not a kernel module", ctx.OtherModuleName(m)))
			}
		})

	return
}

func (m *KernelModule) generateKbuildArgs(ctx blueprint.ModuleContext) map[string]string {
	var extraIncludePaths []string

	g := getBackend(ctx)

	extraCflags := m.Build().BuildProps.Cflags

	for _, includeDir := range m.Build().BuildProps.Include_dirs {
		includeDir = "-I" + filepath.Join(g.sourcePrefix(), includeDir)
		extraIncludePaths = append(extraIncludePaths, includeDir)
	}

	kmod_build := filepath.Join(bobdir, "scripts", "kmod_build.py")
	kdir := m.Properties.Build.Kernel_dir

	extraSymbols := ""
	if len(m.extraSymbolsFiles(ctx)) > 0 {
		extraSymbols = "--extra-symbols " + strings.Join(m.extraSymbolsFiles(ctx), " ")
	}

	kbuildOptions := ""
	if len(m.Build().Kbuild_options) > 0 {
		kbuildOptions = "--kbuild-options " + strings.Join(m.Build().Kbuild_options, " ")
	}

	return map[string]string{
		"kmod_build":           kmod_build,
		"extra_includes":       strings.Join(extraIncludePaths, " "),
		"extra_cflags":         strings.Join(extraCflags, " "),
		"kbuild_extra_symbols": extraSymbols,
		"kernel_dir":           kdir,
		"kernel_compiler":      m.Properties.Build.Kernel_compiler,
		"kbuild_options":       kbuildOptions,
		"make_args":            strings.Join(m.Properties.Build.Make_args, " "),
		"output_module_dir":    filepath.Join(m.OutputDir(g), ctx.ModuleDir()),
	}
}

func (m *KernelModule) GenerateBuildActions(ctx blueprint.ModuleContext) {
	if IsEnabled(m) {
		ctx.Config().(*Config).Generator.KernelModuleActions(m, ctx)
	}
}

func KernelModuleFactory(config *Config) (blueprint.Module, []interface{}) {
	module := &KernelModule{}

	module.Properties.Features.Init(config, reflect.TypeOf(BuildProps{}))
	module.Properties.Build.Target.Init(config, reflect.TypeOf(BuildProps{}))
	module.Properties.Build.Host.Init(config, reflect.TypeOf(BuildProps{}))

	return module, []interface{}{&module.Properties, &module.SimpleName.Properties}
}
