#!/bin/bash
#
# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2018 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------

set -e

function update_bplist {
    local TEMP_LIST_FILE=$(mktemp)

    # First identify the build output directories and format the arguments
    # that will exclude them in the main find command.
    local EXCLUDE=()
    while read -d '' -r ; do
        EXCLUDE+=(-o -path "${REPLY%/.out-dir}")
    done < <(find . -mindepth 2 -maxdepth 3 -type f -name ".out-dir" -print0)

    # Locate all build.bp under the current directory. Exclusions:
    # * hidden directories (starting with .)
    # * the bob directory
    # * Bob build directories (these contain a file .out-dir)
    #   For reasonable efficiency, only look for build dirs at the top level
    find . -mindepth 1 \
     -type d \( -name ".*" -o -name bob "${EXCLUDE[@]}" \) -prune \
     -o -name build.bp -print > "${TEMP_LIST_FILE}"
    echo ./bldsys/bob/Blueprints >> "${TEMP_LIST_FILE}"
    echo ./bldsys/bob/blueprint/Blueprints >> "${TEMP_LIST_FILE}"

    sort "${TEMP_LIST_FILE}" -o "${TEMP_LIST_FILE}"

    if cmp -s "${BLUEPRINT_LIST_FILE}" "${TEMP_LIST_FILE}"; then
        rm "${TEMP_LIST_FILE}"
    else
        mv "${TEMP_LIST_FILE}" "${BLUEPRINT_LIST_FILE}"
    fi
}

# Check the user isn't trying to execute this script directly
if [[ ! -h "${BASH_SOURCE[0]}" ]] ; then
    echo "$0 must be executed from the build directory created by bootstrap_linux.bash"
    exit 1
fi

# Switch to the build directory
cd $(dirname "${BASH_SOURCE[0]}")

BOOTSTRAP=".bob.bootstrap"
source "${BOOTSTRAP}"

# Check for missing configuration
if [ ! -f mali.config ] ; then
    echo "mali.config is missing. Use config or menuconfig to configure the DDK."
    exit 1
fi

# Move to the working directory
cd "${WORKDIR}"

# Check for deprecated configuration
"${BOB_DIR}/config_system/warn_on.py" -d "${SRCDIR}/Mconfig" \
  "${BUILDDIR}/mali.config" -w DEPRECATED -w UNRELEASED ${BOB_CONFIG_OPTS}

# The work directory happens to be the source directory at the moment.
# If this changes bplist generation will need to be tweaked.
update_bplist

# Build
"${BUILDDIR}/bob" "$@"

# Validate if any objects are already out of date. This implies that the build is incomplete.
# This check is here temporarily to make sure we don't break dependencies.
# Skip if a dry run (-n), running a tool (-t), or running verbose (-v)
# Skip if a phony target is specified, because this throws false positives
if echo "$*" | grep -q -E -v "\-(n|t|v)" ; then
    # Find out the name of targets from ninja, and create a regular expression
    BAD_TARGETS=`ninja -f "${BUILDDIR}/build.ninja" -t targets rule phony`
    BAD_TARGETS=`echo ${BAD_TARGETS} | sed -e 's? ?\\\\b\\|?g'`

    if echo " $*" | grep -q -E -v " (${BAD_TARGETS})" ; then
        ninja -f "${BUILDDIR}/build.ninja" -d explain -n "$@" > ${BUILDDIR}/depcheck.log 2>&1
        if grep -q -e "is dirty" ${BUILDDIR}/depcheck.log ; then
            echo "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@"
            echo "@       A DEPENDENCY HAS FAILED TO BUILD CORRECTLY             @"
            echo "@                                                              @"
            echo "@ Targets were out of date after the build finished - this     @"
            echo "@ should never happen. Details can be found in the log:        @"
            echo "@               <BUILDDIR>/depcheck.log                        @"
            echo "@                                                              @"
            echo "@ Note: This warning can be triggered incorrectly if you       @"
            echo "@       modify source files or directory content (affecting a  @"
            echo "@       glob) during the build. A particular case where this   @"
            echo "@       will happen is the first build in a source directory,  @"
            echo "@       where Python byte compiled files get saved to the      @"
            echo "@       source directory.                                      @"
            echo "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@"
            cat ${BUILDDIR}/depcheck.log
            exit 1
        fi
    fi
fi

# Generate ddk build information in PPA format
"${SRCDIR}/bldsys/gen_ddk_json.py" "${BUILDDIR}/config.json" -o "${BUILDDIR}/ddk.json"
