# -----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as
# authorised by a licensing agreement from ARM Limited
# (C) COPYRIGHT 2010-2012, 2014-2017 ARM Limited
# ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorised
# copies and copies may only be made to the extent permitted
# by a licensing agreement from ARM Limited.
# -----------------------------------------------------------------------------

import sys, platform, os
from utils import info_print
from SCons.Script import Glob

# Canonical mapping of Python's view of world to our view of world
_dBuildOS = {
        "Windows"   : "windows",
        "WindowsCE" : "windows",
        "Linux"     : "linux",
        "Darwin"    : "osx",
        "Symbian"   : "symbian",
        "Android"   : "android",
}

# =============================================================================
def get_build_os():
	rHostOS = platform.system()
	try:
		rBuildOS = _dBuildOS[rHostOS]
	except KeyError:
		sys.stderr.write('ERROR: No host OS support [%s]' % rBuildOS)
		exit( 1 )

	return rBuildOS

# =============================================================================
def get_appropriate_os( sEnv ):

	if( sEnv['os'] == 'linux' ):
		# Check we can build for linux on the host platform
		host = sEnv['PLATFORM']
		if host != 'posix':
			sys.stderr.write('*** Cannot build for linux on %s\n' % host)
			exit(1)
		else:
			info_print(sEnv, "Configuring build for Linux.")
			return _OSLinux()

	elif( sEnv['os'] == 'osx' ):
		# Check we can build for osx on the host platform
		host = sEnv['PLATFORM']
		if host != 'darwin':
			sys.stderr.write('*** Cannot build for osx on %s\n' % host)
			exit(1)
		else:
			info_print(sEnv, "Configuring build for OSX.")
			return _OSX()

	elif( sEnv['os'] == 'android' ):
		# Check we can build for Android on the host platform
		host = sEnv['PLATFORM']
		if host != 'posix':
			sys.stderr.write('*** Cannot build for Android on %s\n' % host)
			exit(1)
		else:
			info_print(sEnv, "Configuring build for Android.")
			osAndroid = _OSAndroid()
                        if( sEnv['standalone'] ):
                            osAndroid.dCCDefines.update( {'ANDROID' : 1} )
                        return osAndroid

	else:
		sys.stderr.write('*** os "%s" is not supported \n' % (sEnv['os']))
		exit(1)


# =============================================================================
class _OSBase( object ):


	# -------------------------------------------------------------------------
	def __init__( self ):

		self.dCCDefines = dict()
		self.lLDFlags   = list()
		self.rKernel    = None
		self.yTarget    = False

	# -------------------------------------------------------------------------
	def set_windowing_system( self, lWinSys ):

		if ( lWinSys[0] =='dummy' ):
			self.dCCDefines.update( {'EGL_DUMMY' : 1} )
		self.yBuildDummy = 'dummy' in lWinSys

	# -------------------------------------------------------------------------
	def set_as_target( self ):
		self.yTarget = True

	# -------------------------------------------------------------------------
	def update_environment( self, sEnv ):

		sEnv.Append( CPPDEFINES =self.dCCDefines )
		sEnv.Append( LINKFLAGS  =self.lLDFlags )
		if not self.rKernel:
			self.rKernel = sEnv['os']
		sEnv.AppendUnique( kernel = self.rKernel, delete_existing=1 )

# =============================================================================
class _OSUnix( _OSBase ):


	# -------------------------------------------------------------------------
	def __init__( self ):

		_OSBase.__init__( self )

	# -------------------------------------------------------------------------
	def set_windowing_system( self, lWinSys ):

		# The EGL_* defines are used for two reasons: including the correct
		# khronos headers and for differentiating winsys implementations in
		# tests.
		# Both of these initially affect the 'default' winsys (i.e. the one
		# used when EGL_EXT_platform is *not* used), so set the defines based
		# on that. Setting two of the defines would just cause chaos.
		if ( lWinSys[0] =='fbdev' ):
			self.dCCDefines.update( { 'EGL_FBDEV' : 1} )
		elif ( lWinSys[0] == 'wayland'):
			self.dCCDefines.update( { 'EGL_WAYLAND' : 1} )
		elif ( lWinSys[0] == 'gbm'):
			self.dCCDefines.update( { 'EGL_GBM'   : 1} )

		_OSBase.set_windowing_system(self, lWinSys)

# =============================================================================
class _OSLinux( _OSUnix ):


	# -------------------------------------------------------------------------
	def __init__( self ):

		_OSBase.__init__( self )
		self.rKernel = 'linux'
		self.dCCDefines.update( {'_GNU_SOURCE' : 1} )

	# -------------------------------------------------------------------------
	def set_windowing_system( self, lWinSys ):
		# Record if this build includes GBM or Wayland. We can't rely
		# on the EGL_GBM defines, as they only signify the default winsys.
		self.yBuildGBM = 'gbm' in lWinSys
		self.yBuildWayland = 'wayland' in lWinSys

		_OSUnix.set_windowing_system(self, lWinSys)

	# -------------------------------------------------------------------------
	def add_thirdparty_lib( self, sEnv, winsys, added_thirdparty_winsys_dirs_list ):

		# quit if we have already added this lib and make sure it is valid
		if winsys in added_thirdparty_winsys_dirs_list or not winsys in ('gbm', 'wayland', 'vulkan_wsi'):
			return

		# record adding this lib so we don't add it again
		added_thirdparty_winsys_dirs_list.append(winsys)

		if Glob('#'+winsys+'/thirdparty'):
			sEnv.Append( CPPPATH='#'+winsys+'/thirdparty/include' )
			if sEnv["arch"] == "arm_v8" and winsys == "wayland":
				sEnv.Append( LINKFLAGS = '-L'+winsys+'/thirdparty/lib64' )
				sEnv.Append( LINKFLAGS = '-Wl,-rpath-link,'+winsys+'/thirdparty/lib64' )
			else:
				sEnv.Append( LINKFLAGS = '-L'+winsys+'/thirdparty/lib' )
				sEnv.Append( LINKFLAGS = '-Wl,-rpath-link,'+winsys+'/thirdparty/lib' )


	# -------------------------------------------------------------------------
	def check_and_add_lib_drm( self, sEnv, winsys, added_thirdparty_list ):
		lib_path = '#' + winsys + "/thirdparty/include/libdrm"
		if Glob(lib_path):
			self.add_thirdparty_lib( sEnv, winsys, added_thirdparty_list )
			sEnv.Append( CPPPATH=lib_path )
			return True
		return False

	# -------------------------------------------------------------------------
	def update_environment( self, sEnv ):
		_OSBase.update_environment( self, sEnv)

		if( self.yTarget and (sEnv['os'] != 'android') and not sEnv.GetOption('clean') ):

			# keep track of the thirdparty libs we add
			# this means we can easily avoid adding the libs twice
			added_thirdparty_winsys_dirs_list = list()

			# include libdrm
			# first try to get the one from the winsys thirdparty dir
			# otherwise fall back to any thirdparty dir with libdrm
			if self.yBuildGBM and self.check_and_add_lib_drm(sEnv, 'gbm', added_thirdparty_winsys_dirs_list ):
				pass
			elif self.yBuildWayland and self.check_and_add_lib_drm(sEnv, 'wayland', added_thirdparty_winsys_dirs_list ):
				pass
			# now take libdrm from anywhere starting with gbm
			elif self.check_and_add_lib_drm(sEnv, 'gbm', added_thirdparty_winsys_dirs_list ):
				pass
			elif self.check_and_add_lib_drm(sEnv, 'wayland', added_thirdparty_winsys_dirs_list ):
				pass
			elif self.check_and_add_lib_drm(sEnv, 'vulkan_wsi', added_thirdparty_winsys_dirs_list ):
				pass

			# add the thirdparty libs we need (they may already have been added for libdrm)
			if self.yBuildGBM:
				self.add_thirdparty_lib( sEnv, 'gbm', added_thirdparty_winsys_dirs_list )

			if self.yBuildWayland or int(sEnv['wayland_server']):
				self.add_thirdparty_lib( sEnv, 'wayland', added_thirdparty_winsys_dirs_list )

			# Don't bother checking for pkg-config if a different
			# pkg-config command or path haven't been specified - otherwise
			# we will just be checking host system libraries
			if ( sEnv['ENV']['PKG_CONFIG'] != 'pkg-config' or
						'PKG_CONFIG_PATH' in sEnv['ENV'] ):
				sConf = sEnv.GetConfigure()

				if 'sysroot' in sEnv:
					sEnv['ENV'].setdefault('PKG_CONFIG_SYSROOT_DIR', sEnv['sysroot'])

				if not sConf.CheckPkgConfig():
					print '*******************************************************************'
					print 'WARNING: pkg-config cannot be found, and required third party'
					print '         packages are not included in the DDK source tree.'
					print '         See the accompanying Integration Manual for details.'
					print '         Compilation failures are likely to occur.'
					print '*******************************************************************'

				# The packages themselves are checked lazily, to only check the packages we
				# care about. See add_ext_libs_to_env in utils.py

				sConf.Finish()

# =============================================================================
class _OSX( _OSUnix ):


	# -------------------------------------------------------------------------
	def __init__( self ):

		_OSBase.__init__( self )

		#For Mac OS X, the head file <sys/types.h> require _DARWIN_C_SOURCE
		#to be defined to enable u_int, u_long ...... types
		self.dCCDefines.update( {'_DARWIN_C_SOURCE' : 1} )

		#For Mac OS X, add the Application Services framework
		self.lLDFlags.append ( '-Wl,-framework,ApplicationServices' )

		#For Mac OS X, add I/O Kit framework
		self.lLDFlags.append ( '-Wl,-framework,IOKit' )

# =============================================================================
class _OSAndroid( _OSUnix ):


	# -------------------------------------------------------------------------
	def __init__( self ):

		_OSBase.__init__( self )
		self.rKernel = 'linux'
		self.dCCDefines.update( {'_GNU_SOURCE' : 1} )

	def update_environment( self, sEnv ):
		_OSBase.update_environment( self, sEnv)
