# -----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as
# authorised by a licensing agreement from ARM Limited
# (C) COPYRIGHT 2010-2018 ARM Limited
# ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorised
# copies and copies may only be made to the extent permitted
# by a licensing agreement from ARM Limited.
# -----------------------------------------------------------------------------

import sys, os, commands, subprocess, re
from utils import info_print

# =============================================================================
def get_gcc_version( rCC ):
	# --dumpversion returns only major number in newer versions of gcc; -dumpfullversion works like -dumpversion in
	# older versions of gcc, but those older versions don't support -dumpfullversion...
	sProcessHandle = subprocess.Popen(["%s -dumpfullversion" % rCC ], shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
	rGCC_stdout = sProcessHandle.stdout.readline()
	sProcessHandle.stdout.read()
	sProcessHandle.wait()
	if 0 != sProcessHandle.returncode: # try old version of version option
		sProcessHandle = subprocess.Popen(["%s -dumpversion" % rCC ], shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
		rGCC_stdout = sProcessHandle.stdout.readline()
		sProcessHandle.stdout.read()
		sProcessHandle.wait()
	if 0 == sProcessHandle.returncode:
		lTokens = rGCC_stdout.split(".")
		iMajor = int(lTokens[0])
		iMinor = int(lTokens[1])
		# Some GCC versions return M.ma, some M.ma.x, and some M.ma.mb
		# If we can't decode the 3rd digit safely - treat it as 0
		try:
			iPatch = int(lTokens[2])
			return (iMajor, iMinor, iPatch)
		except:
			return (iMajor, iMinor, 0)
	return (-1, -1, -1) # both version calls failed

# =============================================================================
def get_appropriate_toolchain( sEnv ):

	if( sEnv['os'] in ['linux', 'osx', 'android'] ):

		dOptimize = dict()
		if( sEnv['arch'].find('x86')  != -1):
			if( sEnv['toolchain']  == 'clang'):
				info_print(sEnv, "Configure build toolchain LLVM/clang.\n")
				return _ToolchainClang( sEnv['arch'], sEnv['os'], dOptimize )

			info_print(sEnv, "Configure build toolchain Gnu.\n")
			return _ToolchainGnu( sEnv['arch'], sEnv['os'], dOptimize )

		elif( sEnv['arch'].find('arm') != -1):

			if( 'core' in sEnv ):
				dOptimize['core'] = sEnv['core']

			if( sEnv['toolchain']  == 'rvct'):
				info_print(sEnv, "Configure build toolchain RVCT.\n")
				return _ToolchainRVCT( sEnv['arch'], sEnv['os'], dOptimize )
			elif( sEnv['toolchain']  == 'clang'):
				info_print(sEnv, "Configure build toolchain LLVM/clang.\n")
				return _ToolchainClang( sEnv['arch'], sEnv['os'], dOptimize )

			info_print(sEnv, "Configure build toolchain Gnu.\n")
			return _ToolchainGnu( sEnv['arch'], sEnv['os'], dOptimize )
		else:
			sys.stderr.write('*** "%s" is not supported"\n' % (sEnv['arch']))
			exit(1)

	else:
		sys.stderr.write('*** "%s" is not supported"\n' % (sEnv['os']))
		exit(1)

# =============================================================================
def set_proper_ld_flags( sToolChain ):

	# for x86 builds query the host CXX linker for the path to the libstdc++ shared library
	if "x86" in sToolChain.rTargetArch:
		command_string = sToolChain.rCXX + ' -print-file-name=libstdc++.so ' + sToolChain.dTargetArch[sToolChain.rTargetArch]
		# MIDCL-1885 - need to get the path at the end of the output.  Tools such as
		# BullsEyeCoverage intercept prefix the output with a copyright string.
		output = os.path.dirname(os.path.abspath(commands.getoutput(command_string))).split()
		libstdcxx_dir = output[-1]
		sToolChain.lLDFlags.append('-Wl,-rpath,{0}'.format(libstdcxx_dir))
		if ( 'linux' == sToolChain.rTargetOS ) :
			sToolChain.lLDFlags.append('-Wl,--enable-new-dtags')

	# Determine linker parameters needed by the current 'ld' in the $PATH
	ld_path=sToolChain.rTOOLPREFIX + 'ld'
	sProcessHandle = subprocess.Popen([ld_path + ' --version'], shell=True,
									  stderr=subprocess.PIPE, stdout=subprocess.PIPE)
	rLD_stdout = sProcessHandle.stdout.readline()

	sProcessHandle.stdout.read()
	sProcessHandle.wait()

	# Force using bfd linker in Chrome, which uses the gold linker. This should not be done for android, as android builds
	# work with the gold linker, and switching to the bfd linker causes issues when building with LTO.
	if rLD_stdout.count('gold') == 1 and sToolChain.rTargetOS != "android":
		sToolChain.lLDFlags.append('-fuse-ld=bfd')

	if 0 == sProcessHandle.returncode:
		use_NO_AS_NEEDED = 0
		use_COPY_DT_NEEDED = 0

		# GNU LD version strings are of one of the following formats:
		# (where <junk> is text we don't care about)
		# GNU ld version 2.x.y.z <junk>
		# GNU ld (<junk>) 2.x.y.z <junk>
		ld_version = re.match("GNU ld (?:version )?(?:\([^)]*\) )?([0-9.]+)", rLD_stdout)

		if ld_version is None:
			sys.stderr.write("*** Unrecognised version of ld: %s\n" % rLD_stdout)
			# Fall back to assuming the flags will be needed
			use_NO_AS_NEEDED = 1
			use_COPY_DT_NEEDED = 1
		elif ld_version.group(1) in ["2.20","2.21"]:
			use_NO_AS_NEEDED = 1
		elif ld_version.group(1) >= "2.22":
			use_NO_AS_NEEDED = 1
			use_COPY_DT_NEEDED = 1

		if use_NO_AS_NEEDED:
			sToolChain.lLDFlags.append('-Wl,--no-as-needed')
		if use_COPY_DT_NEEDED:
			# in 2.22 --no-copy-dt-needed-entries became the default
			sToolChain.lLDFlags.append('-Wl,--copy-dt-needed-entries')

# =============================================================================
def set_compiler_debug_flags( sToolChain, rSymbols, rSanitizer, sEnv ):

		if( rSymbols == '1'):
			sToolChain.lCCFlags.append( '-g' )
			sToolChain.lASFlags.append( '-g' )
			if ( sToolChain.rTargetArch == 'arm_v8' ):
				# arm_v8 has a bug where it needs this flag to produce correct stack frames.
				sToolChain.lCCFlags.append( '-mno-omit-leaf-frame-pointer' )
			sToolChain.lCCFlags.append( '-fno-omit-frame-pointer' )
			# optimize sibling calls controls tail function return optimizations that
			# cannot preserve the stack frame. 
			sToolChain.lCCFlags.append( '-fno-optimize-sibling-calls' )
			if( rSanitizer == 'address'):
				sToolChain.lCCFlags.append( '-fsanitize=address' )
				sToolChain.lLDFlags.append( '-fsanitize=address' )
			elif( rSanitizer == 'undefined'):
				sToolChain.lCCFlags.append( '-fsanitize=undefined' )
				sToolChain.lCCFlags.append( '-fsanitize=float-cast-overflow' )
				sToolChain.lLDFlags.append( '-fsanitize=undefined' )
		else:
			sToolChain.lCCFlags.append( '-fomit-frame-pointer' )
			if (rSanitizer != 'none'):
				sys.stderr.write("bldsys error: Using sanitizer checkers requires symbol information to be present. Use with symbols=1 only.\n")
				sys.exit( -1 )

# =============================================================================
def set_compiler_optimization_flags( sToolChain, sEnv ):

	optCCFlags = []
	optLDFlags = []
	optASFlags = []

	if ( sToolChain.rTargetArch == 'arm_v7'):

		optCCFlags.append( '-mfloat-abi=' + sEnv['floatabi'] )
		optCCFlags.append( '-fno-math-errno' )
		optCCFlags.append( '-fno-signed-zeros' )

		if( int( sEnv["simd"] ) ):
			optCCFlags.append( '-mfpu=neon' )
			optASFlags.append( '-mfpu=neon' )
			optLDFlags.append( '-mfpu=neon' )
		else:
			optCCFlags.append( '-mfpu=vfpv3' )
			optASFlags.append( '-mfpu=vfpv3' )

		if ('1' == sEnv['thumb'] ):
			optCCFlags.append( '-mthumb' )
		else:
			optCCFlags.append( '-marm' )

	if ('1' == sEnv['optimize'] ):
		if( sEnv['flavour'] == 'small' ):
			optCCFlags.append( '-Os' )
			optLDFlags.append( '-Os' )
		elif( sEnv['flavour'] == 'fast' ):
			optCCFlags.append( '-O3' )
	else:
		optCCFlags.append( '-O0' )

	if( 'core' in sToolChain.dOptimize ):
		rCore = sToolChain.dOptimize['core']
		optCCFlags.append( '-mtune=%s' % rCore )
		optLDFlags.append( '-mtune=%s' % rCore )

	if ( sToolChain.rTargetArch == 'arm_v7'):
		optCCFlags.append( '-mfloat-abi=' + sEnv['floatabi'] )
		optCCFlags.append( '-fno-math-errno' )
		optCCFlags.append( '-fno-signed-zeros' )

		if( int( sEnv["simd"] ) ):
			optCCFlags.append( '-mfpu=neon' )
			optASFlags.append( '-mfpu=neon' )
			optLDFlags.append( '-mfpu=neon' )
		else:
			optCCFlags.append( '-mfpu=vfpv3' )
			optASFlags.append( '-mfpu=vfpv3' )

	if sEnv['debug'] == '1':
		sToolChain.lSHCCFlags.append( '-fvisibility=default' )
		optCCFlags.append( '-funwind-tables' )
	else:
		sToolChain.lSHCCFlags.append( '-fvisibility=hidden' )

	sToolChain.lCCFlags.extend(optCCFlags)
	sToolChain.lLDFlags.extend(optLDFlags)
	sToolChain.lASFlags.extend(optASFlags)

	sEnv['lto_cflags'] = []
	sEnv['lto_ldflags'] = []
	if (sEnv['lto_requested'] == '1'):
		# If LTO has been requested, it is important to pass
		# proper optimization flags to the final invocation of
		# compiler that is used to link the binary.
		sEnv['lto_ldflags'].extend( optCCFlags )

# =============================================================================

class _ToolchainBase( object ):

	"""
	This class implements a base class from which other toolchains are derived.
	"""

	# -------------------------------------------------------------------------
	def __init__( self, rTargetArch, dOptimize=dict() ):

		self.rTargetArch = rTargetArch
		self.lCFlags     = list()
		self.lCCFlags    = list()
		self.lCXXFlags   = list()
		self.lASFlags    = list()
		self.lLDFlags    = list()
		self.lSHCCFlags  = list()
		self.dOptimize   = dOptimize

	# -------------------------------------------------------------------------
	def update_environment( self, sEnv ):

		sEnv.Append( CFLAGS     = self.lCFlags )
		sEnv.Append( CCFLAGS    = self.lCCFlags )
		sEnv.Append( SHCCFLAGS  = self.lSHCCFlags )
		sEnv.Append( CXXFLAGS   = self.lCXXFlags )
		sEnv.Append( LINKFLAGS  = self.lLDFlags )
		sEnv.Append( ASFLAGS    = self.lASFlags )
		if( 'MALI_GCC_WORKAROUND_MIDCOM_6165' not in sEnv['CPPDEFINES'] ):
			sEnv.Append( CPPDEFINES = {'MALI_GCC_WORKAROUND_MIDCOM_6165' : '0'} )
		if( 'MALI_GCC_WORKAROUND_GRAPHICSSW_4542' not in sEnv['CPPDEFINES'] ):
			sEnv.Append( CPPDEFINES = {'MALI_GCC_WORKAROUND_GRAPHICSSW_4542' : '0'} )
		if( 'MALI_GCC_WORKAROUND_GRAPHICSSW_7663' not in sEnv['CPPDEFINES'] ):
			sEnv.Append( CPPDEFINES = {'MALI_GCC_WORKAROUND_GRAPHICSSW_7663' : '0'} )
		if( 'MALI_GCC_WORKAROUND_GRAPHICSSW_9291' not in sEnv['CPPDEFINES'] ):
			sEnv.Append( CPPDEFINES = {'MALI_GCC_WORKAROUND_GRAPHICSSW_9291' : '0'} )
		if int(sEnv['release']) == 0:
		    sEnv.Append( CCFLAGS = ['-Wundef', '-Wno-error=undef'] )

# =============================================================================
class _ToolchainGnu( _ToolchainBase ):

	"""
	This class implements an instance of the toolchain for Gnu.
	"""

	# Shared lists for all gcc-like compilers
	DEFAULT_C_FLAGS = [
		'-std=c99',
		'-Wstrict-prototypes'
	]

	DEFAULT_CC_FLAGS = [
		'-pedantic',
		'-Wall',
		'-Wextra',
		'-Wno-long-long',
		'-Wno-variadic-macros',
		'-Wno-overlength-strings',
		'-Wno-unused',
		'-Wno-unused-parameter',
		'-Wno-attributes',
		'-Wno-type-limits',
		'-Wno-missing-field-initializers',
		'-Wformat',
		# Treat all warnings as errors
		'-Werror',
		# Assume pthread support
		'-pthread'
	]

	DEFAULT_CXX_FLAGS = [
		'-fno-rtti',
		# LLVM-3.5 requires compilation with C++11
		'-std=c++11'
	]

	DEFAULT_LD_FLAGS = [
		# Assume pthread support
		'-pthread',
		'-Wl,--allow-shlib-undefined'
	]
	# -------------------------------------------------------------------------
	def __init__( self, rTargetArch, rTargetOS, dOptimize=dict() ):

		_ToolchainBase.__init__(self, rTargetArch, dOptimize)

		self.dTargetArch = {
			'arm_v7'   : '-march=armv7-a',
			'arm_v8'   : '-march=armv8-a',
			'x86_32'   : '-m32',
			'x86_64'   : '-m64'
		}

		# !!! Must be first in line - to be prepended by -Wa for ASPP.
		# Don't use an executable stack for unmarked asm
		self.lASFlags.append( '--noexecstack' )

		self.lCCFlags.append( self.dTargetArch[rTargetArch] )
		self.lASFlags.append( self.dTargetArch[rTargetArch] )
		self.lLDFlags.append( self.dTargetArch[rTargetArch] )

		# Force march to new core to enable atomic builtins
		if( "x86_32" == rTargetArch ):
			self.lCCFlags.append( '-march=pentium' )

		self.lCFlags += self.DEFAULT_C_FLAGS
		self.lCCFlags += self.DEFAULT_CC_FLAGS
		self.lCXXFlags += self.DEFAULT_CXX_FLAGS
		self.lLDFlags += self.DEFAULT_LD_FLAGS

		# Add extra flags specific to 'g++'
		if ( rTargetOS != 'android' ): # on Android we use 'clang' on some modules even with the 'gcc' toolchain
			self.lCXXFlags.append( '-fno-gnu-unique' ) # Do not create STB_GNU_UNIQUE symbols in .so files

		self.rCC     = 'gcc'
		self.rAS     = 'as'
		self.rAR     = 'ar'
		self.rCXX    = 'g++'
		self.rRANLIB = 'ranlib'
		self.rTOOLPREFIX = ''
		self.rTargetArch = rTargetArch
		self.rTargetOS   = rTargetOS
		self.rToolChainLTO = False

		# we should keep these defaults to match what most people use
		if ( self.rTargetArch == 'arm_v7' ):
			self.set_toolchain_prefix( 'arm-linux-gnueabihf-' );
		elif ( self.rTargetArch == 'arm_v8' ):
			self.set_toolchain_prefix( 'aarch64-linux-gnu-' );

	def configure_linker( self, sEnv ):
		if( sEnv['sanitizer'] != 'none'):
			self.rLD = self.rCC
		set_proper_ld_flags( self )

	# -------------------------------------------------------------------------
	def set_debug_symbol_flags( self, rSymbols, rSanitizer, sEnv ):
		set_compiler_debug_flags( self, rSymbols, rSanitizer, sEnv )

	# -------------------------------------------------------------------------
	def set_optimization_flags( self, sEnv ):

		set_compiler_optimization_flags( self, sEnv )

		if ('1' == sEnv['optimize'] ):
			if( sEnv['flavour'] == 'fast' ):
				self.lCCFlags.append( '-Wno-array-bounds' )

			self.lCCFlags.append( '-fno-strict-aliasing' )
			self.lCXXFlags.append( '-fno-strict-aliasing' )

		if ( self.rTargetArch == 'arm_v7') and (sEnv['thumb'] == '1'):
			self.lCCFlags.append( '-mthumb-interwork' )

		if (sEnv['lto_requested'] == '1'):
			sys.stderr.write("bldsys error: No support for LTO optimization with GNU toolchain.\n")
			sys.exit( -1 )

	# -------------------------------------------------------------------------
	def set_toolchain_prefix( self, rPrefix):

		if ( (self.rTargetArch.find('arm') != -1) and (rPrefix == '') ):
			return
		self.rCC     = rPrefix + 'gcc'
		self.rAS     = rPrefix + 'as'
		self.rCXX    = rPrefix + 'g++'
		if ( self.rToolChainLTO ):
			self.rAR     = rPrefix + 'gcc-ar'
			self.rRANLIB = rPrefix + 'gcc-ranlib'
		else:
			self.rAR     = rPrefix + 'ar'
			self.rRANLIB = rPrefix + 'ranlib'
		self.rTOOLPREFIX = rPrefix

	# -------------------------------------------------------------------------
	def set_toolchain_lto( self, rToolChainLTO):

		self.rCC     = self.rTOOLPREFIX + 'gcc'
		self.rAS     = self.rTOOLPREFIX + 'as'
		self.rCXX    = self.rTOOLPREFIX + 'g++'

		if ( rToolChainLTO ):
			self.rAR     = self.rTOOLPREFIX + 'gcc-ar'
			self.rRANLIB = self.rTOOLPREFIX + 'gcc-ranlib'
		else:
			self.rAR     = self.rTOOLPREFIX + 'ar'
			self.rRANLIB = self.rTOOLPREFIX + 'ranlib'
		self.rToolChainLTO = rToolChainLTO

	# -------------------------------------------------------------------------
	def update_environment( self, sEnv ):
		sEnv.Replace( CC = self.rCC, CXX = self.rCXX, AS = self.rAS, AR = self.rAR, RANLIB = self.rRANLIB, TOOLPREFIX=self.rTOOLPREFIX )

		if sEnv['ccache'] == '1':
			sEnv.Replace(CC = 'ccache ' + sEnv['CC'], CXX = 'ccache ' + sEnv['CXX'])

		# When Linaro GCC 4.8 or 4.9 is used with O2/O3 optimization enabled along with arm-thumb interwork and neon,
		# registers may be incorrectly tracked for some very specific cases.
		#
		# The only known case with 4.9 is cobj_surface_format_info_round_up_strides. Improper registers are passed to
		# "or" instruction and results in incorrect calculation of size.
		#
		# Workaround forces O1 optimization for affected function for GCC 4.9.
		# Please check MIDCOM-6165 for more details.
		#
		# The other case with 4.8 is base_mem_gpu_address when O3 is used. Strange instructions will be generated and
		# cause incorrect result.
		#
		# Workaround force O2 optimisation for affected function for gcc 4.8, and for both thumb or no thumb building.
		# Please check GRAPHICSSW-4542 for more details.
		if sEnv['optimize'] == '1' and sEnv['thumb'] == '1' and sEnv['simd'] == '1':
			sEnv.Append( CPPDEFINES = {'MALI_GCC_WORKAROUND_MIDCOM_6165' : '1'} )
		if sEnv['optimize'] == '1' and sEnv['simd'] == '1':
			sEnv.Append( CPPDEFINES = {'MALI_GCC_WORKAROUND_GRAPHICSSW_4542' : '1'} )

		# Linaro GCC 4.9.x compilers haven been seen to generate incorrect code for some functions when optimizing with O3.
		# In these cases, we use compiler defines to limit the optimization levels of the affected functions.
		#
		# See GRAPHICSSW-7663 and GRAPHICSSW-9291.
		if sEnv['optimize'] == '1':
			sEnv.Append( CPPDEFINES = {'MALI_GCC_WORKAROUND_GRAPHICSSW_7663' : '1'} )

			# GRAPHICSSW-9291 only for 4.9.2 and below on x86 builds:
			if (sEnv['arch'].find('x86') != -1):
				if (self.get_cc_version() <= 492):
					sEnv.Append( CPPDEFINES = {'MALI_GCC_WORKAROUND_GRAPHICSSW_9291' : '1'} )

		if ( 'linux' == self.rTargetOS ) :
			sEnv['_LIBFLAGS'] += ' -Wl,--no-undefined -Wl,-rpath-link,$SHARED_LIB_PATH'
			sEnv['_LIBFLAGS'] += ' -Wl,--gc-sections'
		elif ( 'osx' == self.rTargetOS ) :
			sEnv['_LIBFLAGS'] += ' -Wl,-rpath,$SHARED_LIB_PATH'

		# When linking via compiler driver invocation libm needs to be explicity specified. Don't do this by default as
		# it might hide libmali.so building up a dependency on libm (which we don't want to see ever).
		if( sEnv['sanitizer'] != 'none' ):
			sEnv.Append(LIBS = 'm')

		# Prepend with -Wa to indicate a (single) assembler option.
		sEnv.Replace( ASPPCOM = '$CC -Wa,$ASPPFLAGS $CPPFLAGS $_CPPDEFFLAGS $_CPPINCFLAGS -c -o $TARGET $SOURCES' )
		sEnv.Append( SHLINKFLAGS = ['${"-Wl,-soname,"+str(TARGET.file)+"."+str(soversion[0]) if soversion else ""}'] )
		_ToolchainBase.update_environment( self, sEnv )

	# -------------------------------------------------------------------------
	def get_cc_version( self ):
		(iMajor, iMinorA, iMinorB) = get_gcc_version( self.rCC )
		if -1 != iMajor:
			return 100*iMajor + 10*iMinorA + 1*iMinorB

		return -1

# =============================================================================
class _ToolchainRVCT( _ToolchainBase ):

	"""
	This class implements an instance of the toolchain for RVCT.
	"""

	def __init__( self, rTargetArch, rTargetOS, dOptimize=dict() ):
		_ToolchainBase.__init__(self, rTargetArch, dOptimize)

		# Call armcc in GCC compatability mode:
		#  --arm_linux_config_file Use GNU standard libraries.  Uses a configuration
		#                          that points to a Code Sourcery install in /arm.
		#  --translate_gcc         Translate arguments from their GCC counterparts.
		#  --fpmode=ieee_full      Enable the rounding mode to be set.
		self.rCC     = 'armcc --arm_linux_config_file=$ARMCC_CONFIG --translate_gcc -Warmcc,--fpmode=ieee_full'
		# This is presumably not used as the config file is not set.
		self.rCXX    = 'armcc --translate_g++'
		# RVCT does not provide RANLIB, use ar directly.
		self.rRANLIB = '$AR -s'
		self.rTargetArch = rTargetArch
		self.rTargetOS   = rTargetOS
		self.dTargetArch = {
			'arm_v7'   : '-march=armv7-a',
			'x86_32'   : '-m32',
			'x86_64'   : '-m64'
		}

		self.lCCFlags.append( self.dTargetArch[rTargetArch] )
		self.lLDFlags.append( self.dTargetArch[rTargetArch] )

	def set_debug_symbol_flags( self, rSymbols, rSanitizer, sEnv ):
		if( rSymbols == '1'):
			self.lCCFlags.append( '-g' )
		if( rSanitizer != 'none'):
			print "Sanitizers are not supported for RVCT"

	def set_optimization_flags( self, sEnv ):
		if ('1' == sEnv['optimize'] ):
			self.lCCFlags.append( '-O2' )

	def update_environment( self, sEnv ):
		sEnv['ARMCC_CONFIG'] = sEnv.File('#/bldsys/armcc_configuration')
		sEnv['AR'] = sEnv.File('#/bldsys/armar-wrapper')

		sEnv.Replace( CC = self.rCC, CXX = self.rCXX, RANLIB = self.rRANLIB)

		if sEnv['ccache'] == '1':
			sEnv.Replace(CC = 'ccache ' + sEnv['CC'], CXX = 'ccache ' + sEnv['CXX'])

		if ( 'linux' == self.rTargetOS ) :
			sEnv['_LIBFLAGS'] += ' -Wl,--no-undefined -Wl,-rpath-link,$SHARED_LIB_PATH'

		_ToolchainBase.update_environment( self, sEnv )

	def set_toolchain_prefix( self, rPrefix):
		# For now do not allow a prefix
		print "Ignoring prefix {0} for RVCT".format(rPrefix)

	def set_toolchain_lto( self, lto ):
		# For now do not allow a prefix
		print "Ignoring lto for RVCT"

	def configure_linker ( self, sEnv ):
		pass

# =============================================================================
class _ToolchainClang( _ToolchainBase ):

	"""
	This class implements an instance of the toolchain for LLVM/clang.
	"""

	def __init__( self, rTargetArch, rTargetOS, dOptimize=dict() ):
		_ToolchainBase.__init__(self, rTargetArch, dOptimize)

		self.dTargetArch = {
			'x86_32'   : '-m32',
			'x86_64'   : '-m64',
			'arm_v7'   : '-march=armv7-a',
			'arm_v8'   : '-march=armv8-a',
		}

		self.rCC     = 'clang'
		self.rCXX    = 'clang++'
		self.rLD     = 'clang++'
		self.rTOOLPREFIX = ''
		self.rTargetArch = rTargetArch
		self.rTargetOS   = rTargetOS
		if rTargetOS == 'osx':
			# On macOS the binary tools are from LLVM by default
			self.rAS     = 'as'
			self.rAR     = 'ar'
			self.rRANLIB = 'ranlib'
		else:
			self.rAS     = 'llvm-as'
			self.rAR     = 'llvm-ar'
			self.rRANLIB = 'llvm-ranlib'

		self.lASFlags.append( '-Wa,--noexecstack' )

		# Android 'clang' has its own architecture setting
		if rTargetOS != 'android':
			self.lCCFlags.append( self.dTargetArch[rTargetArch] )
			self.lLDFlags.append( self.dTargetArch[rTargetArch] )

		self.lCFlags += _ToolchainGnu.DEFAULT_C_FLAGS
		self.lCCFlags += _ToolchainGnu.DEFAULT_CC_FLAGS
		self.lCXXFlags += _ToolchainGnu.DEFAULT_CXX_FLAGS

		# Add extra silencing specific to 'clang'
		self.lCCFlags.append( '-Wno-tautological-constant-out-of-range-compare' )
		self.lCCFlags.append( '-Wno-duplicate-decl-specifier' )
		self.lCCFlags.append( '-Wno-extended-offsetof' )
		self.lCCFlags.append( '-Wno-format-pedantic' )
		self.lCCFlags.append( '-Wno-unused-parameter' )
		self.lCCFlags.append( '-Wno-varargs' )
		self.lCCFlags.append( '-Wno-gnu-zero-variadic-macro-arguments' )
		self.lCCFlags.append( '-Wno-gnu-redeclared-enum' )
		self.lCCFlags.append( '-Wno-newline-eof' )

		# silence warnings not available in current clang version
		self.lCCFlags.append( '-Wno-unknown-warning-option' )

		# Add extra silencing in case of Android, since parts of the
		# framework would fail the -Werror otherwise when included
		if rTargetOS == 'android':
			self.lCCFlags.append( '-Wno-zero-length-array' )
			self.lCCFlags.append( '-Wno-c11-extensions' )
			self.lCCFlags.append ('-Wno-gnu-include-next' )

	def set_debug_symbol_flags( self, rSymbols, rSanitizer, sEnv ):
		set_compiler_debug_flags( self, rSymbols, rSanitizer, sEnv )

	def set_optimization_flags( self, sEnv ):
		set_compiler_optimization_flags ( self, sEnv )

		optCCFlags = []
		optASFlags = []
		optLDFlags = []

		if( 'core' in self.dOptimize ):
			rCore = self.dOptimize['core']
			optCCFlags.append( '-mtune=%s' % rCore )
			optLDFlags.append( '-mtune=%s' % rCore )

		if ( self.rTargetArch == 'arm_v7'):
			optCCFlags.append( '-mfloat-abi=' + sEnv['floatabi'] )
			optCCFlags.append( '-fno-math-errno' )
			optCCFlags.append( '-fno-signed-zeros' )

			if sEnv["simd"] == '1':
				optCCFlags.append( '-mfpu=neon' )
				optASFlags.append( '-mfpu=neon' )
				optLDFlags.append( '-mfpu=neon' )
			else:
				optCCFlags.append( '-mfpu=vfpv3' )
				optASFlags.append( '-mfpu=vfpv3' )

		if sEnv['debug'] == '1':
			self.lSHCCFlags.append( '-fvisibility=default' )
			optCCFlags.append( '-funwind-tables' )
		else:
			self.lSHCCFlags.append( '-fvisibility=hidden' )

		if sEnv['lto_requested'] == '1':
			sEnv['lto_cflags'].extend(['-flto'])
			sEnv['lto_ldflags'].extend(['-flto','-fuse-ld=gold'])
			# If LTO has been requested, it is important to pass
			# proper optimization flags to the final invocation of
			# compiler that is used to link the binary.
			sEnv['lto_ldflags'].extend( optCCFlags )

		self.lCCFlags.extend( optCCFlags )
		self.lLDFlags.extend( optLDFlags )
		self.lASFlags.extend( optASFlags )

	def update_environment( self, sEnv ):
		sEnv.Replace( CC = self.rCC, CXX = self.rCXX, AS = self.rAS,
					  AR = self.rAR, RANLIB = self.rRANLIB, LD = self.rLD, TOOLPREFIX=self.rTOOLPREFIX )

		if ( 'linux' == self.rTargetOS ) :
			sEnv['_LIBFLAGS'] += ' -Wl,--gc-sections -Wl,--no-undefined -Wl,-rpath-link,$SHARED_LIB_PATH'
		elif ( 'osx' == self.rTargetOS ) :
			sEnv['_LIBFLAGS'] += ' -Wl,-rpath,$SHARED_LIB_PATH'
		sEnv.Append( SHLINKFLAGS = ['${"-Wl,-soname,"+str(TARGET.file)+"."+str(soversion[0]) if soversion else ""}'] )

		# When linking via compiler driver invocation libm needs to be explicity specified. Don't do this by default as
		# it might hide libmali.so building up a dependency on libm (which we don't want to see ever).
		if( sEnv['sanitizer'] != 'none' ):
			sEnv.Append(LIBS = 'm')

		_ToolchainBase.update_environment( self, sEnv )

	def set_toolchain_prefix( self, rPrefix):
		self.rTOOLPREFIX = rPrefix

		# Use CC and CXX for target
		# Assume host toolchain has empty prefix
		if len(rPrefix) > 0:
			if os.environ.has_key('CC'):
				self.rCC = os.environ['CC']
			if os.environ.has_key('CXX'):
				self.rCXX = os.environ['CXX']

		if self.rTargetOS == 'android':
			# Android 'clang' sets the architecture flags internally
			return

		if ( self.rTargetArch == 'arm_v7'):
			self.lCCFlags.append( '-target' )
			self.lCCFlags.append( 'armv7a-linux-gnueabihf' )
			self.lASFlags.append( '-target' )
			self.lASFlags.append( 'armv7a-linux-gnueabihf' )
			self.lLDFlags.append( '-target' )
			self.lLDFlags.append( 'armv7a-linux-gnueabihf' )
		elif ( self.rTargetArch == 'arm_v8'):
			self.lCCFlags.append( '-target' )
			self.lCCFlags.append( 'aarch64-linux-gnu' )
			self.lASFlags.append( '-target' )
			self.lASFlags.append( 'aarch64-linux-gnu' )
			self.lLDFlags.append( '-target' )
			self.lLDFlags.append( 'aarch64-linux-gnu' )

		if len(rPrefix) > 0:
			sysroot_out = commands.getoutput( rPrefix+'gcc -print-sysroot' ).split()[-1]
			self.lCCFlags.append( '--sysroot='+sysroot_out )
			self.lLDFlags.append( '--sysroot='+sysroot_out )
			self.lLDFlags.append( '-B'+sysroot_out+'/../bin' )
			# We determine the cross-gcc specific bits here. The last line in 'gcc -v' usually looks like:
			# gcc version 4.8.3 20140401 (prerelease) (crosstool-NG linaro-1.13.1-4.8-2014.04 - Linaro GCC 4.8-2014.04)
			# of which we are interested only in the '4.8.3' entry
			cross_gcc_ver = commands.getoutput( rPrefix+'gcc -v' ).splitlines()[-1].split()[2]
			self.lLDFlags.append( '-B'+sysroot_out+'/../../lib/gcc/'+rPrefix[:-1]+'/' + cross_gcc_ver )
			self.lLDFlags.append( '-L'+sysroot_out+'/../../lib/gcc/'+rPrefix[:-1]+'/' + cross_gcc_ver )
			self.lLDFlags.append( '-L'+sysroot_out+'/../lib' )
			self.lLDFlags.append( '-Wl,-rpath-link,'+sysroot_out+'/../lib' )
			# Cross-compiling C++ is still done with libstdc++.so, which comes from gcc, and it needs
			# the corresponding header files from the sysroot path
			self.lCXXFlags.append( '-I'+sysroot_out+'/../include/c++/'+cross_gcc_ver )
			self.lCXXFlags.append( '-I'+sysroot_out+'/../include/c++/{0}/{1}'.format(cross_gcc_ver,rPrefix[:-1]) )

	def configure_linker ( self, sEnv ):
		set_proper_ld_flags( self )

