# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2016-2018 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------
#

import os
import sys
import logging
import subprocess
import re
import tempfile

from config_system import get_config, set_config

logger = logging.getLogger(__name__)

def check_exit_code(command, dir=None):
    '''
    Executes the command, while making sure the executable is found in the $PATH,
    and returns the exit code. If the executable wasn't found, returns 1.
    The 'command' needs to be an array of arguments.
    '''
    try:
        output = subprocess.check_output(command, cwd=dir, stderr=subprocess.STDOUT).strip()
    except OSError:
        return 1
    except subprocess.CalledProcessError as e:
        logger.info("Determining exit code: %s" % str(e))
        logger.info(e.output)
        return e.returncode

    return 0

def has_compiler_flag(compiler, flag, extra_flags=None):
    '''
    Return 1 when the compiler supports the flag, or 0 when it does not.
    '''
    with tempfile.NamedTemporaryFile() as fp:
        fp.write("int main() { return 0; }")
        fp.flush()
        outfile = fp.name + ".o"
        cmd = [compiler, "-c", "-x", "c++"]
        if extra_flags and extra_flags != '':
            cmd.extend(extra_flags.split(' '))
        cmd.extend(['-Werror', flag, fp.name, "-o", outfile])
        if check_exit_code(cmd) == 0:
            return 1

        # Clean up the output file.
        try:
            os.remove(outfile)
        except (OSError) as e:
            # Assume compiler did not support the flag if removal fails.
            return 0

    return 0

def compiler_config():
    '''
    Adds runtime-dependant entries by analyzing user-configured flags.
    This is mainly required to allow some configuration-specific flags that are
    impossible or hard to express in Blueprint, because of the need to, say, run
    a binary and get the output from it.
    '''
    host_cxx = ''
    target_cxx = ''
    target_cxx_flags = None
    # No toolchain prefix indicates a native build
    native_build = get_config('TOOLCHAIN_PREFIX')['value'] == ''
    if get_config('TOOLCHAIN_CLANG')['value'] == 'y':
        host_cxx = get_config('CLANGXX_BINARY')['value']
        if native_build:
            target_cxx = host_cxx
        else:
            target_cxx = get_config('CLANGXX_BINARY')['value']
            target_cxx_flags = '-target ' + get_config('CLANG_TARGET')['value']

    elif get_config('TOOLCHAIN_GNU')['value'] == 'y':
        host_cxx = get_config('GXX_BINARY')['value']
        if native_build:
            target_cxx = host_cxx
        else:
            target_cxx = get_config('TOOLCHAIN_PREFIX')['value'] + get_config('GCC_BINARY')['value']
            target_cxx_flags = get_config('GCC_TARGET_FLAGS')['value']

    # The LLVM libraries have no stable C++ API, so -Wnoexcept-type is not useful.
    # But only add -Wno-noexcept-type if the compiler supports it.
    if has_compiler_flag(target_cxx, '-Werror=noexcept-type', target_cxx_flags):
        set_config('EXTRA_LLVM_TARGET_CXXFLAGS', '-Wno-noexcept-type')
    if native_build:
        if has_compiler_flag(host_cxx, '-Werror=noexcept-type'):
            set_config('EXTRA_LLVM_HOST_CXXFLAGS', '-Wno-noexcept-type')


def plugin_exec():
    if get_config('ALLOW_HOST_EXPLORE')['value'] == 'y':
        compiler_config()
