#include <linux/dma-buf.h>
#include <linux/slab.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >=  KERNEL_VERSION(4, 14, 0)
#include <linux/uaccess.h>
#else
#include <asm/uaccess.h>
#endif
#include <asm/cacheflush.h>

#include "mali_venusfb_dmabuf.h"

//#define DMA_BUF_FB_DEBUG

static int dma_buf_fb_attach(struct dma_buf *buf, struct device *dev, struct dma_buf_attachment *attachment)
{
	struct venusfb_info *fbi;
	fbi = buf->priv;

	/* dma_buf is externally locked during call */
	fbi->nr_attached_devices++;
	printk("%s: nr_attached_devices=%d\n", __FUNCTION__, fbi->nr_attached_devices);

	return 0;
}

static void dma_buf_fb_detach(struct dma_buf *buf, struct dma_buf_attachment *attachment)
{
	struct venusfb_info *fbi;
	fbi = buf->priv;

	/* dma_buf is externally locked during call */
	fbi->nr_attached_devices--;
	printk("%s: nr_attached_devices=%d\n", __FUNCTION__, fbi->nr_attached_devices);
}

static struct sg_table *dma_buf_fb_map(struct dma_buf_attachment *attachment, enum dma_data_direction direction)
{
	struct sg_table *sg;
	struct venusfb_info *fbi;
	int ret;
	int nr_pages;

#if LINUX_VERSION_CODE >=  KERNEL_VERSION(4, 8, 0)
    unsigned long attrs = 0;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
	DEFINE_DMA_ATTRS(attrs);
#endif

#ifdef DMA_BUF_FB_DEBUG
	printk("%s: \n", __FUNCTION__);
#endif

	fbi = attachment->dmabuf->priv;

	nr_pages = fbi->video_info.videomemorysizeorder;

#if 0//ndef ARCH_HAS_SG_CHAIN
	/* if the ARCH can't chain we can't have allocs larger than a single sg can hold */
	if (nr_pages > SG_MAX_SINGLE_ALLOC)
		return ERR_PTR(-EINVAL);
#endif

	sg = kmalloc(sizeof(struct sg_table), GFP_KERNEL);
	if (!sg)
		return ERR_PTR(-ENOMEM);

	/* from here we access the allocation object, so lock the dmabuf pointing to it */
	mutex_lock(&attachment->dmabuf->lock);

	// contiguous
	ret = sg_alloc_table(sg, 1, GFP_KERNEL);
	if (ret) {
		mutex_unlock(&attachment->dmabuf->lock);
		kfree(sg);
		return ERR_PTR(ret);
	}

	sg_dma_len(sg->sgl) = nr_pages * PAGE_SIZE;
	sg_set_page(sg->sgl, pfn_to_page(PFN_DOWN((unsigned int)fbi->map_phy_addr)), nr_pages * PAGE_SIZE, 0);
	sg_dma_address(sg->sgl) = (unsigned int)fbi->map_phy_addr;

#if LINUX_VERSION_CODE >=  KERNEL_VERSION(4, 8, 0)
	attrs = DMA_ATTR_SKIP_CPU_SYNC;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
	dma_set_attr(DMA_ATTR_SKIP_CPU_SYNC, &attrs);
#endif

#if LINUX_VERSION_CODE >=  KERNEL_VERSION(4, 8, 0)
	if (!dma_map_sg_attrs(attachment->dev, sg->sgl, sg->nents, direction, attrs)) {
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
	if (!dma_map_sg_attrs(attachment->dev, sg->sgl, sg->nents, direction, &attrs)) {
#endif
		mutex_unlock(&attachment->dmabuf->lock);
		sg_free_table(sg);
		kfree(sg);
		return ERR_PTR(-ENOMEM);
	}

	fbi->nr_device_mappings++;
	mutex_unlock(&attachment->dmabuf->lock);
	return sg;
}

static void dma_buf_fb_unmap(struct dma_buf_attachment *attachment,
							 struct sg_table *sg, enum dma_data_direction direction)
{
	struct venusfb_info *fbi;

#if  LINUX_VERSION_CODE >=  KERNEL_VERSION(4, 8, 0)
	unsigned long attrs = 0;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
	DEFINE_DMA_ATTRS(attrs);
#endif

#ifdef DMA_BUF_FB_DEBUG
	printk("%s: \n", __FUNCTION__);
#endif

	fbi = attachment->dmabuf->priv;

#if LINUX_VERSION_CODE >=  KERNEL_VERSION(4, 8, 0)
	attrs = DMA_ATTR_SKIP_CPU_SYNC;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
	dma_set_attr(DMA_ATTR_SKIP_CPU_SYNC, &attrs);
#endif

#if LINUX_VERSION_CODE >=  KERNEL_VERSION(4, 8, 0)
	dma_unmap_sg_attrs(attachment->dev, sg->sgl, sg->nents, direction, attrs);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
	dma_unmap_sg_attrs(attachment->dev, sg->sgl, sg->nents, direction, &attrs);
#endif
	sg_free_table(sg);
	kfree(sg);

#ifdef CONFIG_ARM64
	/* TODO (MID64-78): The latest version of the Linux kernel (3.11 as of
	 * this writing) assumes full hardware cache coherency on ARM64 and thus
	 * won't explicitly flush any CPU caches as part of DMA unmap. Our test
	 * platform, however, is not cache coherent. We call flush_cache_all() as a
	 * workaround until non-coherent DMA is supported in the ARM64 kernel.
	 */
	flush_cache_all();
#endif

	mutex_lock(&attachment->dmabuf->lock);
	fbi->nr_device_mappings--;
	mutex_unlock(&attachment->dmabuf->lock);
}

static void dma_buf_fb_release(struct dma_buf *buf)
{
#ifdef DMA_BUF_FB_DEBUG
	printk("%s: do nothing \n", __FUNCTION__);
#endif
}

static void dma_buf_fb_mmap_open(struct vm_area_struct *vma)
{
	struct dma_buf *dma_buf;
	struct venusfb_info *fbi;

#ifdef DMA_BUF_FB_DEBUG
	printk("%s: \n", __FUNCTION__);
#endif
	dma_buf = vma->vm_private_data;
	fbi = dma_buf->priv;

	mutex_lock(&dma_buf->lock);
	fbi->nr_cpu_mappings++;
	mutex_unlock(&dma_buf->lock);
}

static void dma_buf_fb_mmap_close(struct vm_area_struct *vma)
{
	struct dma_buf *dma_buf;
	struct venusfb_info *fbi;

#ifdef DMA_BUF_FB_DEBUG
	printk("%s: \n", __FUNCTION__);
#endif
	dma_buf = vma->vm_private_data;
	fbi = dma_buf->priv;

	BUG_ON(fbi->nr_cpu_mappings <= 0);
	mutex_lock(&dma_buf->lock);
	fbi->nr_cpu_mappings--;
	mutex_unlock(&dma_buf->lock);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0)
static int dma_buf_fb_mmap_fault(struct vm_area_struct *vma, struct vm_fault *vmf)
#else
static int dma_buf_fb_mmap_fault(struct vm_fault *vmf)
#endif
{
	struct venusfb_info *fbi;
	struct dma_buf *dmabuf;
	struct page *pageptr;
	int nr_pages;

#ifdef DMA_BUF_FB_DEBUG
	printk("%s: \n", __FUNCTION__);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0)
	dmabuf = vma->vm_private_data;
#else
	dmabuf = vmf->vma->vm_private_data;
#endif
	fbi = dmabuf->priv;

	nr_pages = fbi->video_info.videomemorysizeorder;

	if (vmf->pgoff > nr_pages)
		return VM_FAULT_SIGBUS;

	pageptr = pfn_to_page(PFN_DOWN((unsigned int)fbi->map_vir_addr)) + vmf->pgoff;

	BUG_ON(!pageptr);

	get_page(pageptr);
	vmf->page = pageptr;

	return 0;
}

struct vm_operations_struct dma_buf_fb_vm_ops = {
	.open = dma_buf_fb_mmap_open,
	.close = dma_buf_fb_mmap_close,
	.fault = dma_buf_fb_mmap_fault
};

static int dma_buf_fb_mmap(struct dma_buf *dmabuf, struct vm_area_struct *vma)
{
#ifdef DMA_BUF_FB_DEBUG
	printk("%s: \n", __FUNCTION__);
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0))
	vma->vm_flags |= VM_IO | VM_DONTEXPAND | VM_DONTDUMP;
#else
	vma->vm_flags |= VM_RESERVED | VM_IO | VM_DONTEXPAND;
#endif
	vma->vm_ops = &dma_buf_fb_vm_ops;
	vma->vm_private_data = dmabuf;
	vma->vm_page_prot =  pgprot_writecombine(vm_get_page_prot(vma->vm_flags));

	/*  we fault in the pages on access */

	/* call open to do the ref-counting */
	dma_buf_fb_vm_ops.open(vma);

	return 0;
}

static void *dma_buf_fb_kmap_atomic(struct dma_buf *buf, unsigned long page_num)
{
	/* IGNORE */
#ifdef DMA_BUF_FB_DEBUG
	printk("%s: \n", __FUNCTION__);
#endif
	return NULL;
}

static void *dma_buf_fb_kmap(struct dma_buf *buf, unsigned long page_num)
{
	/* IGNORE */
#ifdef DMA_BUF_FB_DEBUG
	printk("%s: \n", __FUNCTION__);
#endif
	return NULL;
}

static struct dma_buf_ops dma_buf_fb_ops = {
	.attach = dma_buf_fb_attach,
	.detach = dma_buf_fb_detach,
	.map_dma_buf = dma_buf_fb_map,
	.unmap_dma_buf = dma_buf_fb_unmap,
	.release = dma_buf_fb_release,
	.mmap = dma_buf_fb_mmap,

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0)
	.kmap_atomic = dma_buf_fb_kmap_atomic,
	.kmap = dma_buf_fb_kmap
#else
	.map_atomic = dma_buf_fb_kmap_atomic,
	.map = dma_buf_fb_kmap
#endif
};

int mali_venusfb_get_dmabuf(struct fb_info *info, unsigned long arg)
{
	struct fb_dmabuf_export dma_buf_export_data;
	struct venusfb_info *fbi = (struct venusfb_info *) info;
	struct dma_buf *dma_buf;
	int nr_pages;
#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 1, 0))
    struct dma_buf_export_info exp_info;
#endif

#ifdef DMA_BUF_FB_DEBUG
	printk("%s: \n", __FUNCTION__);
#endif
	if (copy_from_user(&dma_buf_export_data, (struct fb_dmabuf_export *)arg, sizeof(dma_buf_export_data)))
		return -EFAULT;

	nr_pages = fbi->video_info.videomemorysizeorder;

#ifdef DMA_BUF_FB_DEBUG
	printk("%s: nr_pages=%d\n", __FUNCTION__, nr_pages);
#endif
#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 1, 0))
    exp_info.owner = THIS_MODULE;
    exp_info.exp_name = __FILE__;
    exp_info.priv = fbi;
    exp_info.ops = &dma_buf_fb_ops;
    exp_info.size = nr_pages * PAGE_SIZE;
    exp_info.flags = 0;
    dma_buf = dma_buf_export(&exp_info);
#else
	dma_buf = dma_buf_export(fbi, &dma_buf_fb_ops, nr_pages * PAGE_SIZE, 0);
#endif

	if (IS_ERR_OR_NULL(dma_buf)) {
		printk("%s: couldn't export dma_buf\n", __FUNCTION__);
		goto no_export;
	}

	dma_buf_export_data.fd = dma_buf_fd(dma_buf, dma_buf_export_data.flags);

	if (dma_buf_export_data.fd < 0) {
		printk("%s: couldn't get fd from dma_buf\n", __FUNCTION__);
		goto no_fd;
	}

	if (copy_to_user((struct fb_dmabuf_export *)arg, &dma_buf_export_data, sizeof(dma_buf_export_data)))
		return -EFAULT;

#ifdef DMA_BUF_FB_DEBUG
	printk("%s: dma_buf_fd=%d\n", __FUNCTION__, dma_buf_export_data.fd);
#endif

	return 0;

no_fd:
	dma_buf_put(dma_buf);
no_export:
	return -EFAULT;
}

